local DEBUG = false
local BOOLS = {}
BOOLS["and"] = true
BOOLS["or"] = true
BOOLS["not"] = true

-- compile a boolean expression to check if
-- keys exists in a table
-- filter("a not b")({a=1}) => true
local function filter(expr)
    local code = string.gsub(expr,
    "(%w+)", function (n)
        if BOOLS[n] then
            return n
        end
        return "(e['" .. n .. "'] ~= nil)"
    end)
    code = "return function(e) return (" .. code .. ") end"
    local fn, err = load(code)
    return fn()
end

local World = class()
function World:init(...)
    self.systems = {...}
    self.entities = {}
    self.newEntities = {}
    self.touches = {}
end
function World:add(...)
    for i, e in ipairs({...}) do
        table.insert(self.newEntities, e)
    end
end
function World:dispatch(action, ...)
    local returnvalue = nil
    for i,system in ipairs(self.systems) do
        local preaction = "pre_"..action
        if system[preaction] then
            system[preaction](system)
        end
        for j,e in ipairs(self.entities) do
            if system[action] ~= nil and system.filter(e) then
                local res = system[action](system, e, ...)
                if res ~= nil then
                    returnvalue = res
                end
                if res == ecs.DELETE then
                    table.remove(self.entities, j)
                end
            end
        end
    end
    return returnvalue
end

function World:update(...)
    if DEBUG then
        fill(233, 80, 219)
        text("entities:" .. #self.entities, 50, HEIGHT-50)
    end
    if #self.newEntities > 0 then
        for i, e in ipairs(self.newEntities) do
            table.insert(self.entities, e)
        end
        self.newEntities = {}
    end
    self:updateNetwork()
    if #self.touches > 0 then
        for i, touch in ipairs(self.touches) do
            self:dispatch("touched", touch)
        end
        self.touches = {}
    end
    return self:dispatch("update", ...)
end

function World:updateNetwork()
    if self.net ~= nil then        
        if self.isHost then
            local data, remoteIp = self.net:receive()
            if data ~= nil then
                if data.action == "subscribe" then
                    self.clients[remoteIp] = true
                elseif data.action == "touches" then
                    for i,t in ipairs(data[1]) do
                        table.insert(self.touches, t)
                    end
                end
            end
            for clientIp,_ in pairs(self.clients) do
                print("send "..clientIp)
                self.net:send(clientIp, {action="updateentities", entities=self.entities})
            end
        else
            if self.serverIp ~= nil then
                if #self.touches > 0 then
                    self.net:send(self.serverIp, {
                        action="touches",
                        self.touches
                    })
                    self.touches = {}
                end
            else
                print("sub")
                self.net:broadcast({action="subscribe"})
            end
            local data, serverIp = self.net:receive()
            if data ~= nil then
                if data.action == "updateentities" then
                    self.entities = data.entities
                    self.serverIp = serverIp
                end
            end
        end
    end
end

function World:touched(touch)
    table.insert(self.touches, {
        pos=touch.pos, state=touch.state,
        isHost=self.isHost
    })
end

function World:enableNetwork(isHost)
    self.net = Net(5099)
    self.isHost = isHost
    self.broadcasttime = ElapsedTime
    self.clients = {}
    self.updatenettimer = everynth(.5)
end

local System = class()
function System:init(expr)
    self.filter = filter(expr)
end

local function tests()
    m = filter("a and b")   
    assert(m({a=1,b=3}) == true)   
end
tests()

ecs = {
    World=World,
    System=System,
    filter=filter,
    DELETE={}
}
